/*
 * crash-manager
 *
 * Copyright (c) 2020 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <assert.h>
#include <fcntl.h>
#include <limits.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <zip.h>
#include <errno.h>
#include <dirent.h>
#include "diagnostics/diagnostics.h"

#define LOG_TAG "CRASH_MANAGER"

#include "defs.h"
#include "shared/log.h"
#include "shared/util.h"
#include "shared/config.h"

#define SHARE_BASE_DIR    "/run/diagnostics/priv/fifo"
#define TMP_FILE_TEMPLATE "file_XXXXXX"

#define JSON_FILE_EXT     ".json"
#define ZIP_FILE_EXT      ".zip"

enum diagnostics_type {
	BT_UNKNOWN,
	BT_FULL_REPORT,
	BT_INFO_JSON,
	BT_DIR
};

static char *type_to_str(enum diagnostics_entry entry) {
	switch (entry) {
	case FULL_REPORT:
		return "FULL_REPORT";
	case INFO_JSON:
		return "INFO_JSON";
	default:
		return "UKNOWN";
	}
}

static char *get_temp_file_name()
{
	char *result;
	if (asprintf(&result, "%s/%s", SHARE_BASE_DIR, TMP_FILE_TEMPLATE) == -1) {
		_E("asprintf() error: %m\n");
		return NULL;
	}

	if (mktemp(result) == NULL || result[0] == '\0') {
		_E("mktemp() error: %m\n");
		free(result);
		return NULL;
	}

	return result;
}

static bool is_report_file(const char *path, const char *reports_dir)
{
	assert(path);
	assert(reports_dir);

	char *r_path = realpath(path, NULL);
	if (r_path == NULL) {
		_E("realpath() for %s error: %m\n", path);
		return false;
	}

	if (access(r_path, F_OK) == -1) {
		_E("access() for %s error: %m\n", path);
		return false;
	}

	char *start = strstr(r_path, reports_dir);

	if (start == NULL) {
		_I("%s is not within %s\n", r_path, reports_dir);
		return false;
	}

	return true;
}

static enum diagnostics_type check_report_type(const char *path)
{
	assert(path);

	if (strstr(path, JSON_FILE_EXT) == &path[strlen(path) - strlen(JSON_FILE_EXT)]) {
		_D("Report type of %s is JSON_FILE\n", path);
		return BT_INFO_JSON;
	}

	if (strstr(path, ZIP_FILE_EXT) == &path[strlen(path) - strlen(ZIP_FILE_EXT)]) {
		_D("Report type of %s is FULL_REPORT\n", path);
		return BT_FULL_REPORT;
	}

	struct stat file_stat;
	if (stat(path, &file_stat) == -1) {
		_E("stat() for %s error: %m\n", path);
		return BT_UNKNOWN;
	}

	if (stat(path, &file_stat) == 0 && S_ISDIR(file_stat.st_mode)) {
		_D("Report type of %s is DIR", path);
		return BT_DIR;
	}

	_I("Report type of %s is UNKNOWN", path);
	return BT_UNKNOWN;
}

static bool ends_with(const char *string, const char *end)
{
	assert(string);
	assert(end);

	ssize_t end_len = strlen(end);
	ssize_t string_len = strlen(string);

	if (end_len == 0 || end_len > string_len)
		return false;

	return strcasecmp(&string[string_len - end_len], end) == 0;
}

static bool file_match(const char *file_name, enum diagnostics_entry entry)
{
	assert(file_name);

	switch(entry) {
	case INFO_JSON:
		return ends_with(file_name, JSON_FILE_EXT);
	default:
		return false;
	}
}

static bool extract_file(zip_t *archive, int index, const char *out_file_name)
{
	bool result = false;
	zip_file_t *zfile = NULL;

	int extracted_fd = open(out_file_name, O_WRONLY | O_CREAT | O_EXCL, S_IRUSR | S_IWUSR);
	if (extracted_fd < 0) {
		_E("open() file \"%s\" for write error: %m\n", out_file_name);
		goto out;
	}

	struct zip_stat stat;
	if (zip_stat_index(archive, index, ZIP_FL_UNCHANGED, &stat) == -1) {
		printf("zip_stat_index() error\n");
		goto out;
	}

	zfile = zip_fopen_index(archive, index, ZIP_FL_UNCHANGED);
	if (zfile == NULL) {
		_E("zip_fopen_index() error\n");
		goto out;
	}

	char buff[0x1000];
	zip_uint64_t sum = 0;

	while (sum < stat.size) {
		zip_int64_t len = zip_fread(zfile, buff, sizeof(buff));
		if (len < 0) {
			_E("zip_fread() error\n");
			goto out;
		}
		if (write(extracted_fd, buff, len) == -1) {
			_E("write() error: %m\n");
			goto out;
		}
		sum += len;
	}

	result = true;
out:
	if (zfile != NULL)
		zip_fclose(zfile);

	if (extracted_fd >= 0)
		close(extracted_fd);

	return result;
}

static int search_file_in_zip(zip_t *archive, enum diagnostics_entry entry)
{
	zip_int64_t count = zip_get_num_entries(archive, ZIP_FL_UNCHANGED);

	for (zip_int64_t i = 0; i < count; i++) {
		const char *entry_name = zip_get_name(archive, i, ZIP_FL_ENC_GUESS);
		if (entry_name == NULL) {
			_E("zip_get_name() error\n");
			continue;
		}

		if (!file_match(entry_name, entry))
			continue;

		return i;
	}
	return -1;
}

static int share_compressed_file(const char *path, enum diagnostics_entry entry)
{
	assert(path);

	int err = 0;
	int fd = -BR_ERR_INTERNAL;

	char *out_file = get_temp_file_name();

	if (out_file == NULL)
		return -BR_ERR_INTERNAL;

	zip_t *archive = zip_open(path, ZIP_RDONLY, &err);
	if (err != 0) {
		_E("zip_open() %s error: %d\n", path, err);
		archive = NULL;
		goto out;
	}

	int file_index = search_file_in_zip(archive, entry);
	if (file_index < 0) {
		_I("Archive \"%s\" doesn't contain file type: %s", path, type_to_str(entry));
		fd = -BR_ERR_PARAM;
		goto out;
	}

	if (extract_file(archive, file_index, out_file)) {
		fd = open(out_file, O_RDONLY);
		if (fd == -1) {
			_E("open() file \"%s\" for read error: %m", out_file);
			goto out;
		}
		_D("share file %s from \"%s\" as \"%s\"", type_to_str(entry), path, out_file);
	}

out:
	if (archive)
		zip_close(archive);
	unlink(out_file);
	free(out_file);
	return fd;
}

static int share_file(const char *path)
{
	assert(path);

	const char *dest_file = get_temp_file_name();

	if (dest_file == NULL)
		return -BR_ERR_INTERNAL;

	_D("share file \"%s\" as \"%s\"", path, dest_file);

	if (copy_file(dest_file, path) == -1) {
		_E("copy_file() error");
		return -BR_ERR_INTERNAL;
	}

	int fd = open(dest_file, O_RDONLY);
	if (fd  == -1) {
		_E("open() file: \"%s\" error: %m\n", path);
		return -BR_ERR_INTERNAL;
	}
	unlink(dest_file);

	return fd;
}

static int share_uncompressed_file(const char *path, enum diagnostics_entry entry)
{
	assert(path);

	struct dirent *ent;
	DIR *dir = NULL;
	int result = -BR_ERR_PARAM;
	if ((dir = opendir(path)) == NULL) {
		_E("opendir() for \"%s\" error: %m\n", path);
		result = -BR_ERR_INTERNAL;
		goto out;
	}

	while((ent = readdir(dir)) != NULL) {
		if (file_match(ent->d_name, entry)) {
			char file_path[PATH_MAX];
			int ret = snprintf(file_path, PATH_MAX, "%s/%s", path, ent->d_name);

			if (ret < 0) {
				_E("sprintf() error: %m");
				result = -BR_ERR_INTERNAL;
				goto out;
			} else if (ret >= PATH_MAX) {
				_E("sprintf(): result string is too long");
				result = -BR_ERR_INTERNAL;
				goto out;
			}
			result = share_file(file_path);
			break;
		}
	}

	if (result < 0)
		_I("Report \"%s\" doesn't contain requested entry type.", path);

out:
	if (dir != NULL)
		closedir(dir);

	return result;
}

static int share_simple_file(const char *path, enum diagnostics_entry entry)
{
	assert(path);

	if (entry != INFO_JSON) {
		_E("Unsupported file type for report \"%s\" (%s)\n", path, type_to_str(entry));
		return -BR_ERR_PARAM;
	}

	return share_file(path);
}

int diagnostics_report_get_file(const char *report_path,
			      const enum diagnostics_entry entry)
{
	assert(report_path);

	config_t config;
	if (!config_init(&config, CRASH_MANAGER_CONFIG_PATH))
		return -BR_ERR_INTERNAL;

	bool is_report = is_report_file(report_path, config.crash_root_path);
	config_free(&config);

	if (!is_report) {
		_E("File \"%s\" doesn't look like a report.", report_path);
		return -BR_ERR_PARAM;
	}

	enum diagnostics_type rtype = check_report_type(report_path);

	switch(rtype) {
	case BT_INFO_JSON:
		return share_simple_file(report_path, entry);
		break;
	case BT_FULL_REPORT:
		if (entry == FULL_REPORT)
			return share_file(report_path);
		return share_compressed_file(report_path, entry);
		break;
	case BT_DIR:
		return share_uncompressed_file(report_path, entry);
		break;
	default:
		_E("%s is an unsupported report file.", report_path);
		return -BR_ERR_PARAM;
	}
}
